import { computed, Ref, ref } from "@vue/composition-api";
import { InMemoryCache } from "apollo-cache-inmemory";
import ApolloClient from "apollo-client";
import { createHttpLink } from "apollo-link-http";
import gql from "graphql-tag";
import { LbAuthService, LbToken } from "./types";

export interface LbAuthServiceOptions {
  graphqlURL: string;
  loginUrl: string;
  tokenName: string;
  exp: number;
  rememberExp: number;
}

export function useLbAuthService (options: LbAuthServiceOptions): LbAuthService {
  const storedRawToken = localStorage.getItem(options.tokenName);
  const storedToken: LbToken = storedRawToken
    ? JSON.parse(storedRawToken)
    : null;

  const authToken: Ref<LbToken> = ref(storedToken || {} as LbToken);
  const _loading = ref(false);
  const _ready = ref(false);

  /*
  https://stackoverflow.com/questions/3562929/how-can-i-get-an-event-to-fire-every-time-localstorage-is-updated-in-safari-5
  After investigating further (and with the help from a friend) I discovered that the storage_handler method is called
  not when the value of a localstorage value changes on the page in my current window or tab, but when it changes in another tab.
  For example, if I have the two tabs open, and have controls in the pages in each tab to change localstorage settings,
  then when I hit the control in the first tab, the storage_handler method is called in the other tab.
  */
  window.addEventListener("storage", (event) => {
    if (event.key === options.tokenName) {
      authToken.value = event.newValue
        ? JSON.parse(event.newValue)
        : {};
    }
  });

  const authorization = computed(() => {
    // NB non è reactive ...new Date().getTime()
    // vale solo la prima volta
    return authToken.value.access_token &&
            (authToken.value.exp + authToken.value.fromTime) > new Date().getTime() // not expired
      ? "Basic " + authToken.value.access_token
      : null;
  });

  const clearToken = async () => {
    authToken.value = {} as LbToken;
    localStorage.removeItem(options.tokenName);
  };

  const logout = async () => {
    clearToken();
    window.location.replace(options.loginUrl);
  };

  const checkToken = async (authorization: string): Promise<boolean> => {
    try {
      _loading.value = true;
      await new ApolloClient({
        cache: new InMemoryCache(),
        link: createHttpLink({
          uri: options.graphqlURL,
          headers: { authorization: authorization }
        })
      }).query({
        query: gql` { __schema { __typename } }`
      });
      return true;
    } catch (error) {
      clearToken();
      if (error.networkError &&
        "statusCode" in error.networkError &&
        error.networkError.statusCode === 401) {
        throw new Error("Username e/o password errata.");
      }
      throw error;
    } finally {
      _loading.value = false;
    }
  };

  const login = async (credentials: { username: string; password: string }, remember: boolean) => {
    const encoded = btoa(credentials.username + ":" + credentials.password);
    // test per verificare che il token generato lato client sia valido
    const validToken = await checkToken("Basic " + encoded);
    if (validToken) {
      const token = {
        access_token: encoded,
        exp: remember ? options.rememberExp : options.exp,
        fromTime: new Date().getTime(),
        username: credentials.username
      } as LbToken;
      authToken.value = token;
      localStorage.setItem(options.tokenName, JSON.stringify(token));
    }
    return validToken;
  };

  const createLoginUrl = () => {
    return options.loginUrl;
  };

  const username = computed(() => authToken.value.username || "");

  if (authorization.value) {
    checkToken(authorization.value)
      .catch(() => console.warn("init token non valido"))
      .finally(() => { _ready.value = true; });
  } else {
    _ready.value = true;
  }

  return {
    ready: computed(() => _ready.value),
    loading: computed(() => _loading.value),
    authorization,
    username,
    createLoginUrl,
    login,
    logout
  };
}
