import { LbAuthService } from "@/core/auth";
import { InjectionKey } from "@vue/composition-api";
import { InMemoryCache, NormalizedCacheObject } from "apollo-cache-inmemory";
import { ApolloClient } from "apollo-client";
import { ApolloLink } from "apollo-link";
import { setContext } from "apollo-link-context";
import { ErrorResponse, onError } from "apollo-link-error";
import { createHttpLink } from "apollo-link-http";
import { GraphQLError } from "graphql";

export type LbGraphqlService = ApolloClient<NormalizedCacheObject>

export const LbGraphqlServiceKey: InjectionKey<LbGraphqlService> = Symbol("LbGraphqlService");

export interface LbGraphqlServiceOptions {
  gqlApplicationUrl: string;
  authService: LbAuthService;
  onAuthorizationError?: () => void;
  onApplicationGrantError?: () => void;
}

export function useLbGraphqlService(options: LbGraphqlServiceOptions): LbGraphqlService {
  const errorLink: ApolloLink = onError((response: ErrorResponse) => {
    if (response.networkError) {
      console.error("[Network error]: " + JSON.stringify(response.networkError));
      if ("statusCode" in response.networkError) {
        const statusCode = response.networkError.statusCode;
        if (statusCode === 401) {
          console.error("Utente non autorizzato");
          if (options.onAuthorizationError) {
            options.onAuthorizationError();
          } else {
            options.authService.logout();
          }
        }
      }
    }
    if (response.graphQLErrors) {
      response.graphQLErrors.forEach((error: GraphQLError) => {
        if (error.extensions) {
          console.error(`[GraphQL error extensions]: ${JSON.stringify(error.extensions)}`);
        } else {
          console.error(`[GraphQL error]: ${JSON.stringify(error)}`);
        }
      });
    }
  });

  // HTTP connection to the API
  const httpLink: ApolloLink = createHttpLink({
    uri: options.gqlApplicationUrl
  });

  const authLink: ApolloLink = setContext((operation, { headers }) => {
    return {
      headers: {
        ...headers,
        authorization: options.authService.authorization.value
      }
    };
  });

  return new ApolloClient({
    link: errorLink.concat(authLink).concat(httpLink),
    cache: new InMemoryCache({
      addTypename: false
    }),
    defaultOptions: {
      watchQuery: {
        fetchPolicy: "no-cache"
      },
      query: {
        fetchPolicy: "no-cache"
      },
      mutate: {
        fetchPolicy: "no-cache"
      }
    }
  });
}
